use core::fmt::{self, Write};

pub type SmallStr = String;

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum PrimitiveType {
    Bool,
    Int8,
    Uint8,
    Int16,
    Uint16,
    Int32,
    Uint32,
    Int64,
    Uint64,
    Float32,
    Float64,
}

impl PrimitiveType {
    fn as_str(self) -> &'static str {
        match self {
            PrimitiveType::Bool => "bool",
            PrimitiveType::Int8 => "int8_t",
            PrimitiveType::Uint8 => "uint8_t",
            PrimitiveType::Int16 => "int16_t",
            PrimitiveType::Uint16 => "uint16_t",
            PrimitiveType::Int32 => "int32_t",
            PrimitiveType::Uint32 => "uint32_t",
            PrimitiveType::Int64 => "int64_t",
            PrimitiveType::Uint64 => "uint64_t",
            PrimitiveType::Float32 => "float32",
            PrimitiveType::Float64 => "float64",
        }
    }

    pub fn size(self) -> usize {
        match self {
            PrimitiveType::Bool => 1,
            PrimitiveType::Int8 => 1,
            PrimitiveType::Uint8 => 1,
            PrimitiveType::Int16 => 2,
            PrimitiveType::Uint16 => 2,
            PrimitiveType::Int32 => 4,
            PrimitiveType::Uint32 => 4,
            PrimitiveType::Int64 => 8,
            PrimitiveType::Uint64 => 8,
            PrimitiveType::Float32 => 4,
            PrimitiveType::Float64 => 8,
        }
    }
}

impl fmt::Display for PrimitiveType {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str(self.as_str())
    }
}

#[derive(Clone, Debug, PartialEq)]
pub struct ArrayType {
    pub element_type: Box<FieldType>,
    pub length: Option<usize>,
}

impl ArrayType {
    pub fn new(element_type: FieldType, length: Option<usize>) -> Self {
        Self {
            element_type: Box::new(element_type),
            length,
        }
    }
}

impl fmt::Display for ArrayType {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.element_type.fmt(f)?;
        f.write_char('[')?;
        if let Some(length) = self.length {
            length.fmt(f)?;
        }
        f.write_char(']')
    }
}

#[derive(Clone, Debug, PartialEq)]
pub enum FieldType {
    Primitive(PrimitiveType),
    String,
    Time,
    Duration,
    Header,
    Array(ArrayType),
    Other(SmallStr),
}

impl fmt::Display for FieldType {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            FieldType::Primitive(primitive) => primitive.fmt(f),
            FieldType::String => "string".fmt(f),
            FieldType::Time => "time".fmt(f),
            FieldType::Duration => "duration".fmt(f),
            FieldType::Header => "Header".fmt(f),
            FieldType::Array(array) => array.fmt(f),
            FieldType::Other(other) => other.fmt(f),
        }
    }
}

#[derive(Clone, Debug, PartialEq)]
pub struct MessageField {
    pub name: SmallStr,
    pub typ: FieldType,
}

impl fmt::Display for MessageField {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{} {}", &self.typ, &self.name)
    }
}

#[derive(Clone, Debug, PartialEq)]
pub struct MessageDefinition {
    pub fields: Vec<MessageField>,
}

impl fmt::Display for MessageDefinition {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        for field in &self.fields {
            writeln!(f, "{}", field)?;
        }
        Ok(())
    }
}

pub struct MessageDefinitions {
    pub primary: MessageDefinition,
    pub dependencies: Vec<(String, MessageDefinition)>,
}